from __future__ import absolute_import, division, print_function

import pytest

from basicco.qualname import qualname, QualnameError


@pytest.mark.parametrize("force_ast", (True, False))
def test_qualname(force_ast):

    # Get qualified name for regular classes.
    assert qualname(X, force_ast=force_ast) == "X"
    assert qualname(X.Y, force_ast=force_ast) == "X.Y"
    assert qualname(X.Y.Z, force_ast=force_ast) == "X.Y.Z"
    assert qualname(X.Y.Z.method, force_ast=force_ast) == "X.Y.Z.method"

    assert qualname(A, force_ast=force_ast) == "A"
    assert qualname(A.Y, force_ast=force_ast) == "A.Y"
    assert qualname(A.Y.C, force_ast=force_ast) == "A.Y.C"
    assert qualname(A.Y.C.method, force_ast=force_ast) == "A.Y.C.method"

    # AST parsing should fail for classes generated on the fly.
    with pytest.raises(QualnameError):
        print(qualname(type("Z", (object,), {"BLA": 1}), force_ast=True))

    # Built-ins.
    assert qualname(int, force_ast=force_ast) == "int"
    assert qualname(float, force_ast=force_ast) == "float"
    assert qualname(bool, force_ast=force_ast) == "bool"


class X(object):
    class Y(object):
        class Z(object):
            def method(self):
                pass


class A(object):
    class Y(object):
        class C(object):
            def method(self):
                pass


if __name__ == "__main__":
    pytest.main()



from __future__ import absolute_import, division, print_function

import pytest
from pytest import main

from basicco.mapping_proxy import MappingProxyType


def test_wrap_dict():
    internal_dict = {"foo": "bar"}
    proxy_dict = MappingProxyType(internal_dict)
    assert proxy_dict["foo"] == "bar"

    with pytest.raises(TypeError):
        proxy_dict["foo"] = "foo"

    with pytest.raises(TypeError):
        del proxy_dict["foo"]

    internal_dict["foo"] = "foo"
    assert proxy_dict["foo"] == "foo"

    del internal_dict["foo"]
    assert "foo" not in proxy_dict


if __name__ == "__main__":
    main()



from __future__ import absolute_import, division, print_function

import pytest

from basicco import caller_module


def test_caller_module():
    def func():
        return caller_module.caller_module()

    assert func() == __name__


if __name__ == "__main__":
    pytest.main()



from __future__ import absolute_import, division, print_function

import pytest
import six

from basicco.explicit_hash import ExplicitHashMeta


def test_force_explicit_hash_declaration():
    class Class(six.with_metaclass(ExplicitHashMeta, object)):
        pass

    with pytest.raises(TypeError):

        class SubClass(Class):
            def __eq__(self, other):
                raise NotImplementedError()

        assert not SubClass

    class SubClass(Class):
        def __hash__(self):
            raise NotImplementedError()

        def __eq__(self, other):
            raise NotImplementedError()

    assert SubClass


if __name__ == "__main__":
    pytest.main()



"""Namespace/Bunch-like object."""

from __future__ import absolute_import, division, print_function

import re
import copy

import six
from tippo import TYPE_CHECKING, Iterable, Tuple, TypeVar

if TYPE_CHECKING:
    from tippo import Any, Mapping, Iterator

__all__ = ["Namespace", "MutableNamespace", "NamespacedMeta"]


_VT = TypeVar("_VT")

_WRAPPED_SLOT = "__wrapped__"
_MEMBER_REGEX = re.compile(r"^[a-zA-Z_]\w*$")
_init = lambda s, w: object.__setattr__(s, _WRAPPED_SLOT, w)
_read = lambda s: object.__getattribute__(s, _WRAPPED_SLOT)


class Namespace(Iterable[Tuple[str, _VT]]):
    """Read-only namespace that wraps a mapping."""

    __slots__ = (_WRAPPED_SLOT,)

    def __init__(self, wrapped=None):
        # type: (Mapping[str, _VT] | Namespace[_VT] | None) -> None
        """
        :param wrapped: Mapping/Namespace to be wrapped.
        """
        if wrapped is None:
            wrapped = {}
        elif isinstance(wrapped, Namespace):
            wrapped = _read(wrapped)
        _init(self, wrapped)

    def __dir__(self):
        keys = {k for k in _read(self) if isinstance(k, str) and not hasattr(type(self), k) and _MEMBER_REGEX.match(k)}
        return sorted(keys)

    def __eq__(self, other):
        return isinstance(other, Namespace) and _read(other) == _read(self)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "{}({})".format(type(self).__name__, _read(self))

    def __reduce__(self):
        return type(self), (_read(self),)

    def __copy__(self):
        return type(self)(_read(self).copy())

    def __deepcopy__(self, memo=None):
        if memo is None:
            memo = {}
        self_id = id(self)
        try:
            deep_copied = memo[self_id]
        except KeyError:
            cls = type(self)
            deep_copied = memo[self_id] = cls.__new__(cls)
            _init(deep_copied, copy.deepcopy(_read(self), memo))
        return deep_copied

    def __len__(self):
        # type: () -> int
        return len(_read(self))

    def __iter__(self):
        # type: () -> Iterator[Tuple[str, _VT]]
        for key, value in _read(self).items():
            yield key, value

    def __contains__(self, name):
        # type: (str) -> bool
        return name in _read(self)

    def __getattribute__(self, name):
        # type: (str) -> _VT
        cls = type(self)
        if hasattr(cls, name):
            return object.__getattribute__(self, name)
        try:
            return _read(self)[name]
        except KeyError:
            raise AttributeError(name)


class MutableNamespace(Namespace[_VT]):
    """Mutable namespace that wraps a mapping."""

    __slots__ = ()

    def __setattr__(self, name, value):
        # type: (str, _VT) -> None
        cls = type(self)
        if hasattr(cls, name):
            member = getattr(cls, name)
            if hasattr(member, "__set__"):
                object.__setattr__(self, name, value)
            else:
                error = "can't change '{!r}' attribute/method".format(name)
                raise AttributeError(error)
        else:
            _read(self)[name] = value

    def __delattr__(self, name):
        # type: (str) -> None
        cls = type(self)
        if hasattr(cls, name):
            member = getattr(cls, name)
            if hasattr(member, "__delete__"):
                object.__delattr__(self, name)
            else:
                error = "can't delete '{!r}' attribute/method".format(name)
                raise AttributeError(error)
        else:
            try:
                del _read(self)[name]
            except KeyError:
                exc = AttributeError(name)
                six.raise_from(exc, None)
                raise exc


class NamespacedMeta(type):
    """Metaclass that adds a private mutable `__namespace__` as a class property."""

    @property
    def __namespace__(cls):
        # type: () -> MutableNamespace[Any]
        """Class namespace."""
        try:
            return cls.__dict__["__namespace"]
        except KeyError:
            namespace = MutableNamespace()  # type: MutableNamespace[Any]
            type.__setattr__(cls, "__namespace", MutableNamespace())
            return namespace



"""Python 2.7 compatible way of getting the qualified name. Inspired by wbolster/qualname."""

from __future__ import absolute_import, division, print_function

import ast
import inspect

import six
from tippo import TYPE_CHECKING

if TYPE_CHECKING:
    from tippo import Callable

__all__ = ["QualnameError", "qualname"]


_cache = {}  # type: dict[str, dict[int, str]]


class QualnameError(Exception):
    """Raised when could not get the qualified name from AST parsing."""


def qualname(obj, fallback=None, force_ast=False):
    # type: (Callable, str | None, bool) -> str
    """
    Try to find out the qualified name for a class or function.
    This function uses AST parsing in Python 2.7 to try to replicate the qualified name functionality available in
    Python 3.3+.

    :param obj: Function or class.
    :param fallback: Value to return if couldn't get qualified name. Will error if None.
    :param force_ast: Force use AST parser to get qualified name from code.
    :return: Qualified name (or None if not available).
    :raises TypeError: Provided object is not a function or a class.
    :raises QualnameError: Could not get qualified name from AST parsing.
    """

    # Not available for non-callables.
    if not callable(obj) or not hasattr(obj, "__name__") or not hasattr(obj, "__module__"):
        error = "can't determine qualified name for instances of {}".format(repr(type(obj).__name__))
        raise TypeError(error)
    obj_name = obj.__name__
    obj_module = obj.__module__  # type: ignore

    # Native qualified name or manually defined.
    if not force_ast:
        try:
            return getattr(obj, "__qualname__")
        except QualnameError:  # in case this function gets called when getting the attr
            if fallback is not None:
                return fallback
            else:
                raise
        except AttributeError:
            pass

    # Try to match the root of the module in case the object is not nested.
    try:
        module = __import__(obj_module, fromlist=[obj_name])
    except ImportError as e:
        error = "couldn't import module {} for {}; {}".format(repr(obj_module), repr(obj_name), e)
        exc = QualnameError(error)
        six.raise_from(exc, None)
        raise exc
    if getattr(module, obj_name, None) is obj:
        return obj_name

    # Get source file name.
    try:
        filename = inspect.getsourcefile(obj)
        if filename is None:
            raise TypeError()
    except TypeError:
        if fallback is None:
            error = "couldn't find source code filename for {}".format(repr(obj_name))
            exc = QualnameError(error)
            six.raise_from(exc, None)
            raise exc
        else:
            return fallback

    # Get line number.
    if inspect.isclass(obj):
        try:
            _, lineno = inspect.getsourcelines(obj)
        except (OSError, IOError):
            if fallback is None:
                error = "source code for class {} could not be retrieved".format(repr(obj_name))
                exc = QualnameError(error)
                six.raise_from(exc, None)
                raise exc
            else:
                return fallback
    elif inspect.isfunction(obj) or inspect.ismethod(obj):
        if hasattr(obj, "im_func"):
            # Extract function from unbound method (Python 2).
            obj = obj.im_func  # type: ignore
        try:
            code = obj.__code__
        except AttributeError:
            code = obj.func_code  # type: ignore
        lineno = code.co_firstlineno
    else:
        if fallback is None:
            error = "source code could not be retrieved for {}".format(repr(obj_name))
            raise QualnameError(error)
        else:
            return fallback

    # Parse the source file to figure out what the qualified name should be.
    if filename in _cache:
        qualified_names = _cache[filename]
    else:
        try:
            with open(filename, "r") as fp:
                source = fp.read()
        except (OSError, IOError) as e:
            if fallback is None:
                error = "could not read source code at {}; {}".format(repr(filename), e)
                exc = QualnameError(error)
                six.raise_from(exc, None)
                raise exc
            else:
                return fallback
        node = ast.parse(source, filename)
        visitor = _Visitor()
        visitor.visit(node)
        qualified_names = _cache[filename] = visitor.qualified_names

    # Sort list of line numbers based on the one found via inspection.
    if qualified_names:
        linenos = sorted(
            qualified_names,
            reverse=True,
            key=lambda k: (k - lineno if k >= lineno else max(qualified_names) + (lineno - k)),
        )
    else:
        linenos = []

    # Iterate over possible line numbers.
    while linenos:
        current_lineno = linenos.pop()

        # Get qualified name from parsing results.
        qualified_name = qualified_names.get(current_lineno, None)
        if qualified_name is None:
            if fallback is None:
                error = ("qualified name could not be retrieved from source code for {}").format(repr(obj_name))
                raise QualnameError(error)
            else:
                return fallback

        # Verify qualified name.
        if qualified_name and "<locals>" not in qualified_name:
            qualified_name_parts = qualified_name.split(".")
            test_obj = module
            while qualified_name_parts:
                part = qualified_name_parts.pop(0)
                try:
                    test_obj = getattr(test_obj, part)
                except AttributeError:
                    break
            else:

                # Extract function from unbound method (Python 2).
                if hasattr(test_obj, "im_func"):
                    test_obj = test_obj.im_func  # type: ignore

                # Return only confirmed to be the same exact object.
                if test_obj is obj and test_obj:
                    return qualified_name

    # No way to reliably retrieve qualified name.
    if fallback is None:
        error = "qualified name could not be retrieved from source code for {}".format(repr(obj.__name__))
        raise QualnameError(error)
    else:
        return fallback


class _Visitor(ast.NodeVisitor):
    def __init__(self):
        self.stack = []
        self.qualified_names = {}

    def store_qualified_name(self, lineno):
        qn = ".".join(n for n in self.stack)
        self.qualified_names[lineno] = qn

    def visit_FunctionDef(self, node):
        self.stack.append(node.name)
        self.store_qualified_name(node.lineno)
        self.stack.append("<locals>")
        self.generic_visit(node)
        self.stack.pop()
        self.stack.pop()

    def visit_ClassDef(self, node):
        self.stack.append(node.name)
        self.store_qualified_name(node.lineno)
        self.generic_visit(node)
        self.stack.pop()



from __future__ import absolute_import, division, print_function

import itertools
import math

import pytest
import six.moves

from basicco.import_path import get_path, import_path, extract_generic_paths


class MyClass(object):
    class MyNestedClass(object):
        pass


def test_import_path():
    assert import_path("math.floor") is math.floor
    assert import_path("itertools.chain") is itertools.chain
    assert import_path("six.moves") is six.moves
    assert import_path("six.moves.collections_abc") is six.moves.collections_abc  # type: ignore

    assert import_path(__name__ + ".MyClass") is MyClass
    assert import_path(__name__ + ".MyClass.MyNestedClass") is MyClass.MyNestedClass

    with pytest.raises(ImportError):
        import_path("module.submodule.<locals>.Test")


def test_get_path():
    assert get_path(math.floor) == "math.floor"
    assert get_path(itertools.chain) == "itertools.chain"
    assert get_path(six.moves) == "six.moves"

    assert get_path(MyClass) == __name__ + ".MyClass"
    assert get_path(MyClass.MyNestedClass) == __name__ + ".MyClass.MyNestedClass"

    class LocalClass(object):
        pass

    with pytest.raises(ImportError):
        get_path(LocalClass)


def test_extract_generic_paths():
    assert extract_generic_paths("Tuple[int, str]") == ("Tuple", ("int", "str"))


if __name__ == "__main__":
    pytest.main()



"""Metaclass that forces `__hash__` to be declared when `__eq__` is declared."""

from __future__ import absolute_import, division, print_function

__all__ = ["ExplicitHashMeta"]


class ExplicitHashMeta(type):
    """Metaclass that forces `__hash__` to be declared when `__eq__` is declared."""

    @staticmethod
    def __new__(mcs, name, bases, dct, **kwargs):
        if "__eq__" in dct and "__hash__" not in dct:
            error = "declared '__eq__' in {!r} but didn't declare '__hash__'".format(name)
            raise TypeError(error)
        return super(ExplicitHashMeta, mcs).__new__(mcs, name, bases, dct, **kwargs)



"""Runtime type checking with support for import paths."""

from __future__ import absolute_import, division, print_function

import itertools

import six
from six import moves
from tippo import TYPE_CHECKING, TypeVar

from .import_path import DEFAULT_BUILTIN_PATHS, import_path

if TYPE_CHECKING:
    from tippo import Any, Callable, Iterable, Type

__all__ = [
    "TEXT_TYPES",
    "type_names",
    "format_types",
    "import_types",
    "is_instance",
    "is_subclass",
    "is_iterable",
    "assert_is_instance",
    "assert_is_subclass",
    "assert_is_callable",
    "assert_is_iterable",
]


_T = TypeVar("_T")
TEXT_TYPES = tuple({str, six.text_type})  # type: tuple[Type, ...]


def format_types(types):
    # type: (Type| str | None | Iterable[Type| str | None]) -> tuple[Type | str, ...]
    """
    Format types into a tuple.

    :param types: Type(s)/Import Path(s).
    :return: Tuple of Types/Import Paths.
    """
    if types is None:
        return (type(None),)
    elif isinstance(types, type) or isinstance(types, six.string_types):
        return (types,)
    elif isinstance(types, moves.collections_abc.Iterable):  # type: ignore
        return tuple(itertools.chain.from_iterable(format_types(t) for t in types))
    else:
        error = "{!r} object is not a valid type".format(type(types).__name__)
        raise TypeError(error)


def type_names(types):
    # type: (Type | str | None | Iterable[Type | str | None]) -> tuple[str, ...]
    """
    Get type names without importing from paths. Can be used for user-feedback purposes.

    :param types: Types.
    :return: Type names.
    """
    return tuple(
        t.split(".")[-1] if isinstance(t, six.string_types) else t.__name__ if isinstance(t, type) else type(t).__name__
        for t in format_types(types)
    )


def import_types(
    types,  # type: Type | str | None | Iterable[Type | str | None]
    builtin_paths=DEFAULT_BUILTIN_PATHS,  # type: Iterable[str]
    generic=True,  # type: bool
):
    # type: (...) -> tuple[Type, ...]
    """
    Import types from lazy import paths.

    :param types: Types.
    :param builtin_paths: Builtin module paths in fallback order.
    :param generic: Whether to import generic.
    :return: Imported types.
    """
    imported_types = tuple(
        import_path(t, builtin_paths=builtin_paths, generic=generic) if isinstance(t, six.string_types) else t
        for t in format_types(types)
    )

    # Expand all text types.
    if len(TEXT_TYPES) > 1:
        text_types = set(TEXT_TYPES)
        if text_types.intersection(imported_types):
            imported_types += tuple(text_types.difference(imported_types))

    # Expand all integer types.
    if len(six.integer_types) > 1:
        integer_types = set(six.integer_types)
        if integer_types.intersection(imported_types):
            imported_types += tuple(integer_types.difference(imported_types))

    return imported_types


def is_instance(
    obj,  # type: Any
    types,  # type: Type | str | None | Iterable[Type | str | None]
    subtypes=True,  # type: bool
    builtin_paths=DEFAULT_BUILTIN_PATHS,  # type: Iterable[str]
    generic=True,  # type: bool
):
    # type: (...) -> bool
    """
    Get whether object is an instance of the provided types.

    :param obj: Object.
    :param types: Types.
    :param subtypes: Whether to accept subtypes.
    :param builtin_paths: Builtin module paths in fallback order.
    :param generic: Whether to import generic.
    :return: True if it is an instance.
    """
    imported_types = import_types(types, builtin_paths=builtin_paths, generic=generic)
    if subtypes:
        return isinstance(obj, imported_types)
    else:
        return type(obj) in imported_types


def is_subclass(
    cls,  # type: Type
    types,  # type: Type | str | None | Iterable[Type | str | None]
    subtypes=True,  # type: bool
    builtin_paths=DEFAULT_BUILTIN_PATHS,  # type: Iterable[str]
    generic=True,  # type: bool
):
    # type: (...) -> bool
    """
    Get whether class is a subclass of the provided types.

    :param cls: Class.
    :param types: Types.
    :param subtypes: Whether to accept subtypes.
    :param builtin_paths: Builtin module paths in fallback order.
    :param generic: Whether to import generic.
    :return: True if it is a subclass.
    :raises TypeError: Did not provide a class.
    """
    if not isinstance(cls, type):
        error = "is_subclass() arg 1 must be a class"
        raise TypeError(error)
    imported_types = import_types(types, builtin_paths=builtin_paths, generic=generic)
    if subtypes:
        return issubclass(cls, import_types(types))
    else:
        return cls in imported_types


def is_iterable(value, include_strings=False):
    # type: (Any, bool) -> bool
    """
    Tell whether a value is an iterable or not.
    By default, strings are not considered iterables.

    :param value: Value.
    :param include_strings: Whether to consider strings as iterables.
    :return: True if iterable.
    """
    return isinstance(value, moves.collections_abc.Iterable) and (  # type: ignore
        (not isinstance(value, six.string_types) or include_strings)
    )


def assert_is_instance(
    obj,  # type: _T
    types,  # type: Type[_T] | str | None | Iterable[Type[_T] | str | None]
    subtypes=True,  # type: bool
    builtin_paths=DEFAULT_BUILTIN_PATHS,  # type: Iterable[str]
    generic=True,  # type: bool
):
    # type: (...) -> _T
    """
    Assert object is an instance of the provided types.

    :param obj: Object.
    :param types: Types.
    :param subtypes: Whether to accept subtypes.
    :param builtin_paths: Builtin module paths in fallback order.
    :param generic: Whether to import generic.
    :raises ValueError: No types were provided.
    :raises TypeError: Object is not an instance of provided types.
    """
    if not is_instance(
        obj,
        types,
        subtypes=subtypes,
        builtin_paths=builtin_paths,
        generic=generic,
    ):
        types = format_types(types)
        if not types:
            error = "no types were provided to perform assertion"
            raise ValueError(error)
        error = "got {!r} object, expected instance of {}{}".format(
            type(obj).__name__,
            ", ".join(repr(n) for n in type_names(types)),
            " or any of {} subclasses".format("their" if len(types) > 1 else "its")
            if subtypes
            else " (instances of subclasses are not accepted)",
        )
        raise TypeError(error)
    return obj


def assert_is_subclass(
    cls,  # type: Type[_T]
    types,  # type: Type[_T] | str | None | Iterable[Type[_T] | str | None]
    subtypes=True,  # type: bool
    builtin_paths=DEFAULT_BUILTIN_PATHS,  # type: Iterable[str]
    generic=True,  # type: bool
):
    # type: (...) -> Type[_T]
    """
    Assert a class is a subclass of the provided types.

    :param cls: Class.
    :param types: Types.
    :param subtypes: Whether to accept subtypes.
    :param builtin_paths: Builtin module paths in fallback order.
    :param generic: Whether to import generic.
    :raises ValueError: No types were provided.
    :raises TypeError: Class is not a subclass of provided types.
    """
    if not is_instance(cls, type, builtin_paths=builtin_paths, generic=generic):
        types = format_types(types)
        error = "got instance of {!r}, expected {}{}{}".format(
            type(cls).__name__,
            "one of " if len(types) > 1 else "class ",
            ", ".join(repr(n) for n in type_names(types)),
            " or any of {} subclasses".format("their" if len(types) > 1 else "its")
            if subtypes
            else " (subclasses are not accepted)",
        )
        raise TypeError(error)

    if not is_subclass(
        cls,
        types,
        subtypes=subtypes,
        builtin_paths=builtin_paths,
        generic=generic,
    ):
        types = format_types(types)
        if not types:
            error = "no types were provided to perform assertion"
            raise ValueError(error)

        error = "got {!r}, expected {}{}{}".format(
            cls.__name__,
            "one of " if len(types) > 1 else "class ",
            ", ".join(repr(n) for n in type_names(types)),
            " or any of {} subclasses".format("their" if len(types) > 1 else "its")
            if subtypes
            else " (subclasses are not accepted)",
        )
        raise TypeError(error)

    return cls


def assert_is_callable(value):
    # type: (Callable[..., _T]) -> Callable[..., _T]
    """
    Assert a value is callable.

    :param value: Value.
    :raises TypeError: Value is not a callable.
    """
    if not callable(value):
        error = "got non-callable {!r} object, expected a callable".format(type(value).__name__)
        raise TypeError(error)
    return value


def assert_is_iterable(value, include_strings=False):
    # type: (Iterable[_T], bool) -> Iterable[_T]
    """
    Assert a value is iterable.
    By default, strings are not considered iterables.

    :param value: Value.
    :param include_strings: Whether to consider strings as iterables.
    :raises TypeError: Value is not iterable.
    """
    if not is_iterable(value, include_strings=include_strings):
        error = "got non-iterable {!r} object, expected an iterable".format(type(value).__name__)
        raise TypeError(error)
    return value



from __future__ import absolute_import, division, print_function

import pytest

from basicco.custom_repr import mapping_repr, iterable_repr


def test_custom_mapping_repr():
    assert (
        mapping_repr(
            {1: 4, "2": 3, 3: 2, "4": "1"},
            prefix="Mapping<",
            template="{key}=={value}",
            separator="; ",
            suffix=">",
            sorting=True,
            sort_key=lambda i: str(i[1]),
            reverse=True,
            key_repr=lambda k: repr("K" + str(k)),
            value_repr=lambda v: str("V" + str(v)),
        )
        == "Mapping<'K1'==V4; 'K2'==V3; 'K3'==V2; 'K4'==V1>"
    )


def test_custom_iterable_repr():
    assert (
        iterable_repr(
            ["a", 1, 2.0, "3.0", 4, None],
            prefix="Iterable -",
            template="> {value} <",
            separator="-",
            suffix="-",
            sorting=True,
            sort_key=lambda v: str(v),
            reverse=True,
            value_repr=lambda v: str("V" + str(v)),
        )
        == "Iterable -> Va <-> VNone <-> V4 <-> V3.0 <-> V2.0 <-> V1 <-"
    )


if __name__ == "__main__":
    pytest.main()



"""Value factoring with support for lazy dot paths."""

from __future__ import absolute_import, division, print_function

import six
from tippo import TYPE_CHECKING

from .import_path import DEFAULT_BUILTIN_PATHS, import_path

if TYPE_CHECKING:
    from tippo import Any, Callable, Iterable, Mapping

__all__ = ["fabricate_value"]


def fabricate_value(factory, value, args=None, kwargs=None, builtin_paths=DEFAULT_BUILTIN_PATHS):
    # type: (str | Callable | None, Any, Iterable | None, Mapping[str, Any] | None, Iterable[str]) -> Any
    """
    Pass value through the factory.

    :param factory: Factory, import path, or None.
    :param value: Input value.
    :param args: Factory arguments.
    :param kwargs: Factory keyword arguments.
    :param builtin_paths: Builtin module paths in fallback order.
    :return: Output value.
    :raises TypeError: Invalid factory type.
    """
    if isinstance(factory, six.string_types):
        factory = import_path(factory, builtin_paths=builtin_paths)
    if factory is None:
        return value
    if not callable(factory):
        error = "{!r} object is not a valid factory".format(type(factory).__name__)
        raise TypeError(error)
    return factory(value, *args or (), **kwargs or {})



"""Import from/generate lazy dot paths."""

from __future__ import absolute_import, division, print_function

import types
import six
from six import moves
from tippo import TYPE_CHECKING

from .qualname import qualname

if TYPE_CHECKING:
    from tippo import Any, Iterable, Type

__all__ = ["DEFAULT_BUILTIN_PATHS", "import_path", "extract_generic_paths", "get_path"]


DEFAULT_BUILTIN_PATHS = (moves.builtins.__name__, "tippo")
_NOTHING = object()
_SPECIAL_VALUES = {
    "...": Ellipsis,
    "None": None,
    "NoneType": type(None),
    "True": True,
    "False": False,
    "NotImplemented": NotImplemented,
    "NotImplementedType": type(NotImplemented),
}
_SPECIAL_PATHS = {v: k for k, v in _SPECIAL_VALUES.items()}


def _import_builtin(path, builtin_paths):
    # type: (str, Iterable[str]) -> Any
    for builtin_path in builtin_paths:
        try:
            return import_path(".".join((builtin_path, path)), builtin_paths=())
        except (ImportError, AttributeError):
            continue
    return _NOTHING


def _get_generic(obj, generic_paths, builtin_paths):
    # type: (Any, tuple[str, ...], Iterable[str]) -> Any
    generics = tuple(import_path(gp, builtin_paths=builtin_paths) for gp in generic_paths)
    if generics:
        if len(generics) == 1:
            return obj[generics[0]]
        else:
            return obj[generics]
    else:
        return obj


def _get_alias_origin(obj):
    # type: (Any) -> tuple[Type | None, tuple | None]
    if getattr(obj, "__origin__", None) is not None and getattr(obj, "__args__", None) is not None:
        return obj.__origin__, obj.__args__
    else:
        return None, None


def import_path(path, builtin_paths=DEFAULT_BUILTIN_PATHS, generic=True):
    # type: (str, Iterable[str], bool) -> Any
    """
    Import from a dot path.

    :param path: Dot path.
    :param builtin_paths: Builtin module paths in fallback order.
    :param generic: Whether to import generic.
    :return: Imported module/object.
    """

    # Special value.
    if path in _SPECIAL_VALUES:
        return _SPECIAL_VALUES[path]

    # Check and split dot path.
    if not path:
        raise ValueError("empty path")
    path, generic_paths = extract_generic_paths(path)
    if not generic:
        generic_paths = ()
    path_parts = path.split(".")

    # Import the module.
    module = _NOTHING
    module_path_parts = []
    for path_part in path_parts:
        module_path_parts.append(path_part)
        current_path = ".".join(module_path_parts)
        try:
            module = __import__(current_path)
        except ImportError:
            module_path_parts.pop()
            break
    if module is _NOTHING:
        builtin_obj = _import_builtin(path, builtin_paths)
        if builtin_obj is not _NOTHING:
            return _get_generic(builtin_obj, generic_paths, builtin_paths)
        error = "could not import module for {!r}".format(path)
        raise ImportError(error)

    # Reset the module path parts (workaround for things like six.moves).
    module_path_parts = module.__name__.split(".")  # type: ignore

    # Import the object.
    obj_path_parts = path_parts[len(module_path_parts) :]
    if not obj_path_parts:
        if generic_paths:
            error = "can't import generics {!r} for module {!r}".format(generic_paths, path)
            raise ImportError(error)
        return module
    previous_obj = module
    obj = _NOTHING
    for path_part in obj_path_parts:
        try:
            obj = previous_obj = getattr(previous_obj, path_part)
        except AttributeError:
            builtin_obj = _import_builtin(path, builtin_paths)
            if builtin_obj is not _NOTHING:
                return _get_generic(builtin_obj, generic_paths, builtin_paths)
            raise

    return _get_generic(obj, generic_paths, builtin_paths)


def extract_generic_paths(path):
    # type: (str) -> tuple[str, tuple[str, ...]]
    """
    Extract generic paths from dot path.

    :param path: Dot path.
    :return: Dot path and generic paths.
    """

    # Check if contains generic brackets.
    if any(t in path for t in ("[", "]")):

        # Separate generic paths string from import path.
        if path.count("[") != path.count("]") or not path.endswith("]"):
            error = "invalid path {!r}".format(path)
            raise ValueError(error)
        first_index = path.index("[") + 1
        stop_index = len(path) - list(reversed(path)).index("]") - 1
        generic_paths = path[first_index:stop_index]

        # Split on commas but not the ones inside brackets.
        extracted_generic_paths = []  # type: list[str]
        current_path = ""
        in_brackets = 0
        for character in generic_paths:
            if character == " ":
                continue
            if character == "[":
                in_brackets += 1
            elif character == "]":
                in_brackets -= 1
            elif character == "," and not in_brackets:
                extracted_generic_paths.append(current_path)
                current_path = ""
                continue
            current_path += character

        # Assert the correct number of levels.
        if in_brackets != 0:
            error = "invalid generic paths {!r} in {!r}".format(generic_paths, path)
            raise ValueError(error)

        # Append the rest of the string.
        if current_path:
            extracted_generic_paths.append(current_path)

        # Separate import path.
        path = path[: first_index - 1]
    else:
        extracted_generic_paths = []

    # Return extracted paths.
    return path, tuple(extracted_generic_paths)


def get_path(
    obj,  # type: Any
    builtin_paths=DEFAULT_BUILTIN_PATHS,  # type: Iterable[str]
    generic=True,  # type: bool
    check=True,  # type: bool
):
    # type: (...) -> str
    """
    Get dot path to an object or module.

    :param obj: Object or module.
    :param builtin_paths: Builtin module paths in fallback order.
    :param generic: Whether to include path to generic as well.
    :param check: Whether to check path for consistency.
    :return: Dot path.
    """

    # Module.
    if isinstance(obj, types.ModuleType):
        return obj.__name__

    # Special paths.
    if obj in _SPECIAL_PATHS:
        return _SPECIAL_PATHS[obj]

    # Get generic suffix.
    generic_suffix = ""
    generic_origin, generic_args = _get_alias_origin(obj)
    if generic_origin is not None:
        assert generic_args is not None
        generic_suffix = "".join(
            (
                "[",
                ", ".join(get_path(ga, builtin_paths=builtin_paths, generic=generic) for ga in generic_args),
                "]",
            )
        )

    # Get qualified name and module.
    try:
        if generic_origin is not None:
            module = generic_origin.__module__
        else:
            module = obj.__module__
    except AttributeError:
        error = "non-importable instance of {!r}".format(type(obj).__name__)
        exc = TypeError(error)  # type: Exception
        six.raise_from(exc, None)
        raise exc
    else:
        if generic_origin is not None:
            short_name = generic_origin.__name__
            qualified_name = qualname(generic_origin, short_name)
        else:
            short_name = obj.__name__
            qualified_name = qualname(obj, short_name)
        if qualified_name.count(".") < 1:
            qualified_name = short_name
    if not module:
        error = "can't get module for {}".format(obj)
        raise AttributeError(error)
    if module in builtin_paths:
        module = ""
    if not qualified_name:
        error = "can't get qualified name for {}".format(obj)
        raise AttributeError(error)
    if "<locals>" in qualified_name:
        error = "local name {!r} is not importable".format(qualified_name)
        raise ImportError(error)

    # Assemble path and check for consistency.
    path = ".".join(p for p in (module, qualified_name) if p)
    if generic and generic_suffix:
        path += generic_suffix
    elif not generic and generic_origin is not None:
        obj = generic_origin
    try:
        imported_obj = import_path(path, builtin_paths=builtin_paths)
        if imported_obj != obj:
            raise ImportError()
    except (ImportError, AttributeError):
        if check:
            error = "import path {!r} is not consistent for {}".format(path, obj)
            exc = ImportError(error)
            six.raise_from(exc, None)
            raise exc

    return path



"""Mapping Proxy type (read-only) for older Python versions."""

from __future__ import absolute_import, division, print_function

__all__ = ["MappingProxyType"]


try:
    from types import MappingProxyType

except ImportError:
    from collections import Mapping  # type: ignore

    class MappingProxyType(Mapping):  # type: ignore
        __slots__ = ("__mapping",)

        def __init__(self, mapping):
            self.__mapping = mapping

        def __getitem__(self, item):
            return self.__mapping[item]

        def __len__(self):
            return len(self.__mapping)

        def __iter__(self):
            for key in self.__mapping:
                yield key



from __future__ import absolute_import, division, print_function

import pytest

from basicco.context_vars import ContextVar


_var = ContextVar("_var")  # type: ContextVar[str]
_default_var = ContextVar("_default_var", default="default")  # type: ContextVar[str]


def test_var():
    with pytest.raises(LookupError):
        _var.get()
    assert _var.get(None) is None

    token = _var.set("foo")
    assert _var.get() == "foo"
    _var.reset(token)

    with pytest.raises(LookupError):
        _var.get()
    assert _var.get(None) is None


def test_default_var():
    assert _default_var.get() == "default"
    assert _var.get(None) is None

    token = _default_var.set("foo")
    assert _default_var.get() == "foo"
    _default_var.reset(token)

    assert _default_var.get() == "default"
    assert _var.get(None) is None


if __name__ == "__main__":
    pytest.main()



"""Rough implementation of ContextVar for Python 2.7 that works with threads."""

from __future__ import absolute_import, division, print_function

import uuid

from tippo import TypeVar, Generic

__all__ = ["ContextVar", "Token"]


_T = TypeVar("_T")


try:
    from contextvars import Token, ContextVar  # noqa
except ImportError:
    from threading import local

    _local = local()
    MISSING = object()

    class Token(Generic[_T]):  # type: ignore
        __slots__ = ("__var", "__old_value")

        MISSING = MISSING

        def __init__(self, var, old_value):
            # type: (ContextVar, _T) -> None
            self.__var = var
            self.__old_value = old_value

        @property
        def var(self):
            # type: () -> ContextVar[_T]
            return self.__var

        @property
        def old_value(self):
            # type: () -> _T
            return self.__old_value

    class ContextVar(Generic[_T]):  # type: ignore
        __slots__ = ("__uuid", "__name", "__default")

        def __init__(self, name, default=MISSING):
            self.__uuid = "_{}".format(str(uuid.uuid4()).replace("-", "_"))
            self.__name = name
            self.__default = default

        def get(self, default=MISSING):
            # type: (...) -> _T
            try:
                return getattr(_local, self.__uuid)
            except AttributeError:
                if default is not MISSING:
                    return default
                if self.__default is not MISSING:
                    return self.__default
            error = "no value set in current context for context variable {!r}".format(self.name)
            raise LookupError(error)

        def set(self, value):
            # type: (_T) -> Token[_T]
            try:
                old_value = getattr(_local, self.__uuid)
            except AttributeError:
                old_value = MISSING
            setattr(_local, self.__uuid, value)
            return Token(self, old_value)  # type: ignore

        def reset(self, token):
            # type: (Token) -> None
            if token.old_value is MISSING:
                delattr(_local, self.__uuid)
            else:
                setattr(_local, self.__uuid, token.old_value)

        @property
        def name(self):
            # type: () -> str
            return self.__name



"""Name mangling utilities."""

from __future__ import absolute_import, division, print_function

import re

__all__ = ["mangle", "unmangle", "extract"]


def mangle(name, cls_name):
    # type: (str, str) -> str
    """
    Mangle a name with a class name if applicable.

    :param name: Name.
    :param cls_name: Class name.
    :return: Mangled name.
    """
    if name.startswith("__") and not name.endswith("__"):
        return "_{}{}".format(cls_name.lstrip("_"), name)
    return name


def unmangle(name, cls_name):
    # type: (str, str) -> str
    """
    Unmangle a mangled name with a class name if applicable.

    :param name: Name.
    :param cls_name: Class name.
    :return: Unmangled name.
    """
    prefix = "_{}".format(cls_name.lstrip("_"))
    if name.startswith("{}__".format(prefix)):
        unmangled_name = name[len(prefix) :]
        if unmangled_name.startswith("__") and unmangled_name.endswith("__"):
            return name
        return name[len(prefix) :]
    return name


def extract(name):
    # type: (str) -> tuple[str, str | None]
    """
    Extract name and class name from mangled name if applicable.

    :param name: Mangled name.
    :return: Unmangled name and class name (or None).
    """
    matches = re.match(r"^_([^_]+)__[^_]+.*?(?<!__)$", name)
    if matches:
        cls_name = matches.groups()[0]
        return name[len(cls_name) + 1 :], cls_name
    return name, None



from __future__ import absolute_import, division, print_function

import pytest

from basicco.fabricate_value import fabricate_value


def test_fabricate_value():
    assert fabricate_value(None, 3) == 3
    assert fabricate_value(str, 3) == "3"
    assert fabricate_value("str", 3) == "3"


if __name__ == "__main__":
    pytest.main()



"""Custom representation functions."""

from __future__ import absolute_import, division, print_function

from tippo import TYPE_CHECKING

if TYPE_CHECKING:
    from tippo import Any, Callable, Hashable, Iterable, Mapping

__all__ = ["mapping_repr", "iterable_repr"]


def mapping_repr(
    mapping,  # type: Mapping
    prefix="{",  # type: str
    template="{key}{value}",  # type: str
    separator=", ",  # type: str
    suffix="}",  # type: str
    sorting=False,  # type: bool
    sort_key=None,  # type: Callable[[tuple[Any, Any]], Any] | None
    reverse=False,  # type: bool
    key_repr=repr,  # type: Callable[[Any], str]
    value_repr=repr,  # type: Callable[[Any], str]
):
    # type: (...) -> str
    """
    Get custom representation of a mapping.

    :param mapping: Mapping.
    :param prefix: Prefix.
    :param template: Item format template ({key} and {value}).
    :param separator: Separator.
    :param suffix: Suffix.
    :param sorting: Whether to sort keys.
    :param sort_key: Sorting key.
    :param reverse: Reverse sorting.
    :param key_repr: Key representation function.
    :param value_repr: Value representation function.
    :return: Custom representation.
    """
    parts = []
    iterable = mapping.items()  # type: Iterable[tuple[Hashable, Any]]
    if sort_key is None:
        sort_key = lambda item: item[0]
    if sorting:
        iterable = sorted(iterable, key=sort_key, reverse=reverse)
    for key, value in iterable:
        part = template.format(key=key_repr(key), value=value_repr(value))
        parts.append(part)
    return prefix + separator.join(parts) + suffix


def iterable_repr(
    iterable,  # type: Iterable
    prefix="[",  # type: str
    template="{value}",  # type: str
    separator=", ",  # type: str
    suffix="]",  # type: str
    sorting=False,  # type: bool
    sort_key=None,  # type: Callable[[Any], Any] | None
    reverse=False,  # type: bool
    value_repr=repr,  # type: Callable[[Any], str]
):
    # type: (...) -> str
    """
    Get custom representation of an iterable.

    :param iterable: Iterable.
    :param prefix: Prefix.
    :param template: Item format template ({key} and {value}).
    :param separator: Separator.
    :param suffix: Suffix.
    :param sorting: Whether to sort the iterable or not.
    :param sort_key: Sorting key.
    :param reverse: Reverse sorting.
    :param value_repr: Value representation function.
    :return: Custom representation.
    """
    parts = []
    if sorting:
        iterable = sorted(iterable, key=sort_key, reverse=reverse)
    for value in iterable:
        part = template.format(value=value_repr(value))
        parts.append(part)
    return prefix + separator.join(parts) + suffix



from __future__ import absolute_import, division, print_function

import pytest

from basicco.recursive_repr import recursive_repr


def test_recursive_repr():
    @recursive_repr
    def my_repr(_self, prefix="MyRepr<", suffix=">"):
        return prefix + my_repr(_self) + suffix

    assert my_repr(object()) == "MyRepr<...>"  # type: ignore


if __name__ == "__main__":
    pytest.main()



from __future__ import absolute_import, division, print_function

import abc
import itertools

import pytest

from basicco.type_checking import (
    assert_is_callable,
    assert_is_instance,
    assert_is_subclass,
    assert_is_iterable,
    is_iterable,
    format_types,
    type_names,
    import_types,
    is_instance,
    is_subclass,
)


class Cls(object):
    pass


class SubCls(Cls):
    pass


cls_path = __name__ + "." + Cls.__name__
subcls_path = __name__ + "." + SubCls.__name__


def test_format_types():
    assert format_types(None) == (type(None),)
    assert format_types((None,)) == (type(None),)
    assert format_types(float) == (float,)
    assert format_types("itertools.chain") == ("itertools.chain",)
    with pytest.raises(TypeError):
        format_types(3)  # type: ignore


def test_type_names():
    assert type_names(None) == (type(None).__name__,)
    assert type_names((None,)) == (type(None).__name__,)
    assert type_names(("module.module.Cls", "abc.abstractmethod", abc.ABCMeta, itertools.chain)) == (
        "Cls",
        "abstractmethod",
        "ABCMeta",
        "chain",
    )


def test_import_types():
    assert import_types(None) == (type(None),)
    assert import_types((None,)) == (type(None),)
    assert import_types(("abc.abstractmethod", abc.ABCMeta, itertools.chain)) == (
        abc.abstractmethod,
        abc.ABCMeta,
        itertools.chain,
    )


def test_is_instance():
    assert is_instance(None, None)
    assert is_instance(None, (None,))
    assert not is_instance(0, None)
    assert not is_instance(0, (None,))

    assert is_instance(Cls(), cls_path) is True
    assert is_instance(Cls(), cls_path, subtypes=False) is True
    assert is_instance(SubCls(), cls_path) is True
    assert is_instance(SubCls(), cls_path, subtypes=False) is False
    assert is_instance(SubCls(), subcls_path, subtypes=False) is True

    assert is_instance(Cls(), (cls_path, Cls, int)) is True
    assert is_instance(Cls(), (cls_path, Cls, int), subtypes=False) is True
    assert is_instance(SubCls(), (cls_path, Cls, int)) is True
    assert is_instance(SubCls(), (cls_path, Cls, int), subtypes=False) is False
    assert is_instance(SubCls(), (subcls_path, SubCls, int), subtypes=False) is True

    assert is_instance(Cls(), (int, float)) is False
    assert is_instance(Cls(), (int, float), subtypes=False) is False
    assert is_instance(SubCls(), (int, float)) is False
    assert is_instance(SubCls(), (int, float), subtypes=False) is False
    assert is_instance(SubCls(), (int, float), subtypes=False) is False

    assert is_instance(Cls(), ()) is False
    assert is_instance(SubCls(), ()) is False


def test_is_subclass():
    assert is_subclass(type(None), None)
    assert is_subclass(type(None), (None,))
    assert not is_subclass(int, None)
    assert not is_subclass(int, (None,))

    assert is_subclass(Cls, cls_path) is True
    assert is_subclass(Cls, cls_path, subtypes=False) is True
    assert is_subclass(SubCls, cls_path) is True
    assert is_subclass(SubCls, cls_path, subtypes=False) is False
    assert is_subclass(SubCls, subcls_path, subtypes=False) is True

    assert is_subclass(Cls, (cls_path, Cls, int)) is True
    assert is_subclass(Cls, (cls_path, Cls, int), subtypes=False) is True
    assert is_subclass(SubCls, (cls_path, Cls, int)) is True
    assert is_subclass(SubCls, (cls_path, Cls, int), subtypes=False) is False
    assert is_subclass(SubCls, (subcls_path, SubCls, int), subtypes=False) is True

    assert is_subclass(Cls, (int, float)) is False
    assert is_subclass(Cls, (int, float), subtypes=False) is False
    assert is_subclass(SubCls, (int, float)) is False
    assert is_subclass(SubCls, (int, float), subtypes=False) is False
    assert is_subclass(SubCls, (int, float), subtypes=False) is False

    assert is_subclass(Cls, ()) is False
    assert is_subclass(SubCls, ()) is False


def test_is_iterable():
    assert not is_iterable(3)
    assert not is_iterable(False)
    assert not is_iterable("foo")

    assert is_iterable("foo", include_strings=True)
    assert is_iterable([])
    assert is_iterable({})
    assert is_iterable(set())


def test_assert_is_instance():
    assert_is_instance(None, None)
    assert_is_instance(None, (None,))
    with pytest.raises(TypeError):
        assert_is_instance(0, None)
        assert_is_instance(0, (None,))

    assert_is_instance(Cls(), cls_path)
    assert_is_instance(Cls(), cls_path, subtypes=False)
    assert_is_instance(SubCls(), cls_path)

    with pytest.raises(TypeError):
        assert_is_instance(SubCls(), cls_path, subtypes=False)

    assert_is_instance(SubCls(), subcls_path, subtypes=False)

    assert_is_instance(Cls(), (cls_path, Cls, int))
    assert_is_instance(Cls(), (cls_path, Cls, int), subtypes=False)
    assert_is_instance(SubCls(), (cls_path, Cls, int))

    with pytest.raises(TypeError):
        assert_is_instance(SubCls(), (cls_path, Cls, int), subtypes=False)

    assert_is_instance(SubCls(), (subcls_path, SubCls, int), subtypes=False)

    with pytest.raises(TypeError):
        assert_is_instance(Cls(), (int, float))

    with pytest.raises(TypeError):
        assert_is_instance(Cls(), (int, float), subtypes=False)

    with pytest.raises(TypeError):
        assert_is_instance(SubCls(), (int, float))

    with pytest.raises(TypeError):
        assert_is_instance(SubCls(), (int, float), subtypes=False)

    with pytest.raises(TypeError):
        assert_is_instance(SubCls(), (int, float), subtypes=False)

    with pytest.raises(ValueError):
        assert_is_instance(Cls(), ())

    with pytest.raises(ValueError):
        assert_is_instance(SubCls(), ())


def test_assert_is_subclass():
    assert_is_subclass(type(None), None)
    assert_is_subclass(type(None), (None,))
    with pytest.raises(TypeError):
        assert_is_subclass(int, None)
        assert_is_subclass(int, (None,))

    assert_is_subclass(Cls, cls_path)
    assert_is_subclass(Cls, cls_path, subtypes=False)
    assert_is_subclass(SubCls, cls_path)

    with pytest.raises(TypeError):
        assert_is_subclass(SubCls, cls_path, subtypes=False)

    assert_is_subclass(SubCls, subcls_path, subtypes=False)

    assert_is_subclass(Cls, (cls_path, Cls, int))
    assert_is_subclass(Cls, (cls_path, Cls, int), subtypes=False)
    assert_is_subclass(SubCls, (cls_path, Cls, int))

    with pytest.raises(TypeError):
        assert_is_subclass(SubCls, (cls_path, Cls, int), subtypes=False)

    assert_is_subclass(SubCls, (subcls_path, SubCls, int), subtypes=False)

    with pytest.raises(TypeError):
        assert_is_subclass(Cls, (int, float))

    with pytest.raises(TypeError):
        assert_is_subclass(Cls, (int, float), subtypes=False)

    with pytest.raises(TypeError):
        assert_is_subclass(SubCls, (int, float))

    with pytest.raises(TypeError):
        assert_is_subclass(SubCls, (int, float), subtypes=False)

    with pytest.raises(TypeError):
        assert_is_subclass(SubCls, (int, float), subtypes=False)

    with pytest.raises(ValueError):
        assert_is_subclass(Cls, ())

    with pytest.raises(ValueError):
        assert_is_subclass(SubCls, ())


def test_assert_is_callable():
    assert_is_callable(int)

    with pytest.raises(TypeError):
        assert_is_callable(3)


def test_assert_is_iterable():
    with pytest.raises(TypeError):
        assert_is_iterable(3)

    with pytest.raises(TypeError):
        assert_is_iterable(False)

    with pytest.raises(TypeError):
        assert_is_iterable("foo")

    assert_is_iterable("foo", include_strings=True)

    assert_is_iterable([])
    assert_is_iterable({})
    assert_is_iterable(set())


if __name__ == "__main__":
    pytest.main()



"""Iterator that yields unique values."""

from __future__ import absolute_import, division, print_function

from tippo import Iterator, TypeVar

__all__ = ["unique_iterator"]


T = TypeVar("T")


def unique_iterator(iterator):
    # type: (Iterator[T]) -> Iterator[T]
    """
    Iterator that yields unique values.

    :param iterator: Iterator.
    :return: Unique iterator.
    """
    seen = set()
    for value in iterator:
        if value in seen:
            continue
        seen.add(value)
        yield value



from __future__ import absolute_import, division, print_function

from pickle import loads, dumps

import six
from pytest import main

from basicco.namespace import _WRAPPED_SLOT, Namespace, MutableNamespace, NamespacedMeta


def test_read_only_namespace():
    wrapped = {"update": "foo"}
    ns = Namespace(wrapped)
    assert hasattr(ns, _WRAPPED_SLOT)
    assert getattr(ns, _WRAPPED_SLOT) is wrapped
    assert dict(ns) == {"update": "foo"}

    assert hasattr(ns, "update")
    assert ns.update == "foo"


def test_pickle():
    wrapped = {"update": "foo"}
    rns = Namespace(wrapped)
    ns = MutableNamespace(wrapped)

    assert loads(dumps(rns)) == rns
    assert loads(dumps(ns)) == ns


def test_namespace():
    wrapped = {"update": "foo"}
    ns = MutableNamespace(wrapped)

    ns.update = "bar"
    assert hasattr(ns, "update")
    assert ns.update == "bar"

    del ns.update
    assert not hasattr(ns, "update")

    ns.update = "foo"
    assert hasattr(ns, "update")
    assert ns.update == "foo"

    ns.update = "bar"
    assert hasattr(ns, "update")
    assert ns.update == "bar"

    del ns.update
    assert not hasattr(ns, "update")


def test_wrap_mutable_namespace():
    mns = MutableNamespace({"foo": "bar"})
    ns = Namespace(mns)
    assert hasattr(mns, "foo")
    assert hasattr(ns, "foo")

    mns.bar = "foo"
    assert hasattr(mns, "bar")
    assert hasattr(ns, "bar")


def test_wrap_namespace():
    ns = Namespace({"foo": "bar"})
    mns = MutableNamespace(ns)
    assert hasattr(ns, "foo")
    assert hasattr(mns, "foo")

    mns.bar = "foo"
    assert hasattr(mns, "bar")
    assert hasattr(ns, "bar")


def test_private_namespace():
    class Class(six.with_metaclass(NamespacedMeta)):
        pass

    assert isinstance(Class.__namespace__, MutableNamespace)

    class SubClassA(Class):
        pass

    assert isinstance(SubClassA.__namespace__, MutableNamespace)

    class SubClassB(Class):
        pass

    assert isinstance(SubClassB.__namespace__, MutableNamespace)

    class SubClassC(SubClassA, SubClassB):
        pass

    assert isinstance(SubClassC.__namespace__, MutableNamespace)

    namespace_ids = {
        id(Class.__namespace__),
        id(SubClassA.__namespace__),
        id(SubClassB.__namespace__),
        id(SubClassC.__namespace__),
    }

    assert len(namespace_ids) == 4


if __name__ == "__main__":
    main()



from __future__ import absolute_import, division, print_function

import pytest
import six

from basicco.runtime_final import _FINAL_CLASS_TAG, _FINAL_METHODS, final, FinalizedMeta


def test_final_class():
    @final
    class Class(six.with_metaclass(FinalizedMeta, object)):
        pass

    # Class should be tagged as final.
    assert getattr(Class, _FINAL_CLASS_TAG) is True

    # Should error when trying to subclass from final class.
    with pytest.raises(TypeError):

        class SubClass(Class):  # type: ignore
            pass

        assert not SubClass


def test_final_method():
    def dummy_decorator(func):
        return func

    class PropertyLikeDescriptor(object):
        def __init__(self, fget):
            self.fget = fget

        def __get__(self, instance, owner):
            return 3

    decorators = [
        dummy_decorator,
        PropertyLikeDescriptor,
        classmethod,
        staticmethod,
        property,
    ]

    # Different kinds of decorated members should be recognized as final.
    for decorator in decorators:

        class Class(six.with_metaclass(FinalizedMeta, object)):
            @decorator
            @final
            def method(self):
                pass

        assert "method" in getattr(Class, _FINAL_METHODS)

        Class.new_method = decorator(final(lambda _: None))
        assert "new_method" in getattr(Class, _FINAL_METHODS)

        del Class.new_method
        assert "new_method" not in getattr(Class, _FINAL_METHODS)

        with pytest.raises(TypeError):

            class SubClass(Class):
                @decorator
                def method(self):  # type: ignore
                    pass

            assert not SubClass


def test_descriptor():
    class Descriptor(object):
        def __init__(self, func):
            self.func = func

        def __get__(self, instance, owner):
            return 3

    class Class(six.with_metaclass(FinalizedMeta, object)):
        @final
        @Descriptor
        def method(self):
            pass

    # Decorated descriptor object should be recognized as final.
    assert "method" in getattr(Class, _FINAL_METHODS)


if __name__ == "__main__":
    pytest.main()



from __future__ import absolute_import, division, print_function

from pytest import main

from basicco.unique_iterator import unique_iterator


def test_unique_iterator():
    assert list(unique_iterator([1, 2, 3, 3, 4, 4, 5])) == [1, 2, 3, 4, 5]


if __name__ == "__main__":
    main()



"""Decorator that prevents infinite recursion for `__repr__` methods."""

from __future__ import absolute_import, division, print_function

import functools

import six
from tippo import TYPE_CHECKING, Counter, TypeVar, overload

from .context_vars import ContextVar

if TYPE_CHECKING:
    from tippo import Callable, Literal

__all__ = ["recursive_repr"]


_T = TypeVar("_T")

_reprs = ContextVar("_reprs")  # type: ContextVar[Counter[int]]


@overload
def recursive_repr(
    maybe_func,  # type: Callable[..., _T]
    max_depth=1,  # type: int | None
    max_repr="...",  # type: str
):
    # type: (...) -> Callable[..., _T]
    pass


@overload
def recursive_repr(
    maybe_func,  # type: Literal[None]
    max_depth=1,  # type: int | None
    max_repr="...",  # type: str
):
    # type: (...) -> Callable[[Callable[..., _T]], Callable[..., _T]]
    pass


def recursive_repr(maybe_func=None, max_depth=1, max_repr="..."):
    """
    Decorate a representation method/function to prevent infinite recursion.

    :param maybe_func: The '__repr__' and/or '__str__' method/function or None.
    :param max_depth: Maximum depth.
    :param max_repr: Repr text to return after max depth has been reached.
    :return: Decorated method function or decorator.
    """

    def decorator(func):
        # type: (Callable[..., _T]) -> Callable[..., _T]

        @functools.wraps(func)
        def decorated(*args, **kwargs):

            # Get self (or cls for class methods).
            try:
                self = args[0]
            except IndexError:
                error = "'recursive_repr' needs to decorate a class/instance method"
                exc = RuntimeError(error)
                six.raise_from(exc, None)
                raise exc
            self_id = id(self)

            # Get reprs counter for current context and increment it for self.
            reprs_token = None
            try:
                reprs = _reprs.get()
            except LookupError:
                reprs = Counter()
                reprs_token = _reprs.set(reprs)
            reprs[self_id] += 1

            # Return representation.
            try:
                if max_depth is not None and reprs[self_id] > max_depth:
                    return max_repr
                else:
                    return func(*args, **kwargs)
            finally:

                # Decrement repr counter and clean up if needed.
                reprs[self_id] -= 1
                if not reprs[self_id]:
                    del reprs[self_id]

                if reprs_token is not None:
                    _reprs.reset(reprs_token)

        return decorated

    # Return decorated or decorated.
    if maybe_func is not None:
        return decorator(maybe_func)
    else:
        return decorator



from __future__ import absolute_import, division, print_function

from pytest import main

from basicco.mangling import mangle, unmangle, extract


def test_mangle():
    assert mangle("bar", "Foo") == "bar"
    assert mangle("_bar", "Foo") == "_bar"
    assert mangle("__bar__", "Foo") == "__bar__"
    assert mangle("__bar", "Foo") == "_Foo__bar"
    assert mangle("__bar", "_Foo") == "_Foo__bar"
    assert mangle("__bar", "__Foo") == "_Foo__bar"


def test_unmangle():
    assert unmangle("bar", "Foo") == "bar"
    assert unmangle("_bar", "Foo") == "_bar"
    assert unmangle("__bar__", "Foo") == "__bar__"
    assert unmangle("_Foo_bar", "Foo") == "_Foo_bar"
    assert unmangle("_Foo__bar__", "Foo") == "_Foo__bar__"
    assert unmangle("_Foo__bar", "Foo") == "__bar"


def test_extract():
    assert extract("bar") == ("bar", None)
    assert extract("_bar") == ("_bar", None)
    assert extract("__bar__") == ("__bar__", None)
    assert extract("_Foo_bar") == ("_Foo_bar", None)
    assert extract("_Foo__bar__") == ("_Foo__bar__", None)
    assert extract("_Foo__bar") == ("__bar", "Foo")


if __name__ == "__main__":
    main()



"""Scrape class for members."""

from __future__ import absolute_import, division, print_function

import inspect

import six
from tippo import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from tippo import Any, Dict, Type, Optional

__all__ = [
    "BASE_OBJECT_MEMBER_NAMES",
    "DEFAULT_MEMBER_FILTER",
    "ALL_MEMBERS_FILTER",
    "MemberFilterProtocol",
    "OverrideFilterProtocol",
    "scrape_class",
]


BASE_OBJECT_MEMBER_NAMES = frozenset(object.__dict__).union({"__dict__", "__module__", "__weakref__"})
DEFAULT_MEMBER_FILTER = lambda b, mn, m: mn not in BASE_OBJECT_MEMBER_NAMES
ALL_MEMBERS_FILTER = lambda b, mn, m: True


class MemberFilterProtocol(Protocol):
    def __call__(self, base, member_name, member):
        # type: (Type, str, Any) -> bool
        pass


class OverrideFilterProtocol(Protocol):
    def __call__(self, base, member_name, member, previous_member):
        # type: (Type, str, Any, Any) -> bool
        pass


def scrape_class(cls, member_filter=DEFAULT_MEMBER_FILTER, override_filter=None):
    # type: (Type, MemberFilterProtocol, Optional[OverrideFilterProtocol]) -> Dict[str, Any]
    """
    Scrape class for specific members.

    :param cls: Class.
    :param member_filter: Member filter function.
    :param override_filter: Member override filter function.
    :return: Dictionary with filtered members mapped by name.
    """
    if override_filter is None:
        override_filter = lambda b, mn, m, pm: member_filter(b, mn, m)

    members = {}  # type: Dict[str, Any]
    for base in reversed(inspect.getmro(cls)):
        for member_name, member in six.iteritems(base.__dict__):
            override = member_name in members
            if (override and override_filter(base, member_name, member, members[member_name])) or (
                not override and member_filter(base, member_name, member)
            ):
                members[member_name] = member
            elif override:
                del members[member_name]

    return members



"""Custom `final` decorator that enables runtime checking for classes and methods."""

from __future__ import absolute_import, division, print_function

import inspect
import functools

from tippo import TYPE_CHECKING, TypeVar, final

if TYPE_CHECKING:
    from tippo import Type, Set, Callable, final

__all__ = ["final", "is_final", "FinalizedMeta"]


_T = TypeVar("_T")
_FINAL_CLASS_TAG = "__is_final_class__"
_FINAL_METHOD_TAG = "__is_final_method__"
_FINAL_METHODS = "__final_methods__"


__final = final


def _final(obj):
    if inspect.isclass(obj):
        if not isinstance(obj, FinalizedMeta):
            error = "class {!r} doesn't have {!r} as its metaclass".format(obj.__name__, FinalizedMeta.__name__)
            raise TypeError(error)
        type.__setattr__(obj, _FINAL_CLASS_TAG, True)
    else:
        object.__setattr__(obj, _FINAL_METHOD_TAG, True)
    return __final(obj)


globals()["final"] = functools.wraps(final)(_final)  # trick IDEs/static type checkers
globals()["final"].__doc__ = "A decorator to indicate final methods and final classes."


def _is_final_member(member):
    _is_final = False

    # Descriptor.
    if hasattr(member, "__get__"):
        _is_final |= getattr(member, _FINAL_METHOD_TAG, False)

        # Has 'fget' getter (property-like).
        if hasattr(member, "fget"):
            _is_final |= getattr(member.fget, _FINAL_METHOD_TAG, False)

    # Static or class method.
    if isinstance(member, (staticmethod, classmethod)):
        _is_final |= getattr(member.__func__, _FINAL_METHOD_TAG, False)

    # Regular method.
    if callable(member):
        _is_final |= getattr(member, _FINAL_METHOD_TAG, False)

    return _is_final


def is_final(obj):
    # type: (Callable) -> bool
    """
    Tell whether a class/method is final.

    :param obj: Class or method.
    :return: True if final.
    """
    if inspect.isclass(obj):
        return getattr(obj, _FINAL_CLASS_TAG, False)
    else:
        return _is_final_member(obj)


class FinalizedMeta(type):
    """Metaclass that enables runtime-checking for `final` decorator."""

    def __init__(cls, name, bases, dct, **kwargs):
        super(FinalizedMeta, cls).__init__(name, bases, dct, **kwargs)
        cls.__gather_final_members()

    def __gather_final_members(cls):

        # Iterate over MRO of the class.
        final_cls = None  # type: Type | None
        final_member_names = set()  # type: Set[str]
        for base in reversed(inspect.getmro(cls)[:-1]):

            # Prevent subclassing final classes.
            if getattr(base, _FINAL_CLASS_TAG, False) is True:
                if final_cls is not None:
                    error = "can't subclass final class {!r}".format(final_cls.__name__)
                    raise TypeError(error)
                final_cls = base

            # Find final members.
            for member_name, member in base.__dict__.items():

                # Can't override final members.
                if member_name in final_member_names:
                    error = "can't override final member {!r}".format(member_name)
                    raise TypeError(error)

                # Keep track of final members.
                if _is_final_member(member):
                    final_member_names.add(member_name)

            # Store final methods.
            type.__setattr__(cls, _FINAL_METHODS, frozenset(final_member_names))

    def __setattr__(cls, name, value):
        super(FinalizedMeta, cls).__setattr__(name, value)
        cls.__gather_final_members()

    def __delattr__(cls, name):
        super(FinalizedMeta, cls).__delattr__(name)
        cls.__gather_final_members()



from __future__ import absolute_import, division, print_function

import pytest

from basicco.scrape_class import scrape_class


class Member(object):
    @staticmethod
    def filter(_base, _member_name, member):
        return isinstance(member, Member)

    @staticmethod
    def override_filter(base, member_name, member, previous_member):
        if not isinstance(member, Member):
            raise TypeError(
                "'{}.{}' is not of type {!r}".format(base.__name__, member_name, type(previous_member).__name__)
            )
        return isinstance(previous_member, Member)


_foo_member = Member()
_bar_member = Member()
_foo_override_member = Member()
_foobar_member = Member()


class BaseA(object):
    a = "a"
    foo = _foo_member


class BaseB(BaseA):
    b = "b"
    bar = _bar_member  # type: Member | None


class BaseC(BaseB):
    c = "c"
    foo = _foo_override_member
    bar = None
    foobar = _foobar_member


def test_default():
    assert scrape_class(object) == {}
    assert scrape_class(BaseA) == {"foo": _foo_member, "a": "a"}
    assert scrape_class(BaseB) == {"foo": _foo_member, "bar": _bar_member, "a": "a", "b": "b"}
    assert scrape_class(BaseC) == {
        "foo": _foo_override_member,
        "bar": None,
        "foobar": _foobar_member,
        "a": "a",
        "b": "b",
        "c": "c",
    }


def test_member_filter():
    assert scrape_class(object, Member.filter) == {}
    assert scrape_class(BaseA, Member.filter) == {"foo": _foo_member}
    assert scrape_class(BaseB, Member.filter) == {"foo": _foo_member, "bar": _bar_member}
    assert scrape_class(BaseC, Member.filter) == {"foo": _foo_override_member, "foobar": _foobar_member}


def test_override_member_filter():
    assert scrape_class(object, Member.filter, Member.override_filter) == {}
    assert scrape_class(BaseA, Member.filter, Member.override_filter) == {"foo": _foo_member}
    assert scrape_class(BaseB, Member.filter, Member.override_filter) == {"foo": _foo_member, "bar": _bar_member}
    with pytest.raises(TypeError) as exc_info:
        scrape_class(BaseC, Member.filter, Member.override_filter)
    assert str(exc_info.value) == "'BaseC.bar' is not of type 'Member'"


if __name__ == "__main__":
    pytest.main()



"""Retrieve the caller's module name."""

from __future__ import absolute_import, division, print_function

import inspect

__all__ = ["caller_module"]


def caller_module(frames=0):
    # type: (int) -> str | None
    """
    Get caller module name if possible.

    :param frames: How many frames to go back.
    :return: Module name or None.
    """
    try:
        frame = inspect.stack()[2 + frames]
        module = inspect.getmodule(frame[0])
    except IndexError:
        return None
    else:
        if module is None:
            try:
                return frame[0].f_globals["__name__"]
            except (IndexError, KeyError):
                return None
        else:
            return module.__name__
